--------------------------------------------------------------------------------
-- Scenario configuration
--------------------------------------------------------------------------------

-- Names for opponent vehicles.
local opponents =
{
	"race_Player1",
	"race_Player2",
	"race_Player3",
	"race_Player4",
	"race_Player5",
}

--------------------------------------------------------------------------------
-- AI configuration
--------------------------------------------------------------------------------

-- Multiplier for AI checkpoint radius, keep this big enough to ensure their
-- progress is tracked correctly
local raceRadiusMultiplier = 1.5

-- Multiplier for AI checkpoint radius, used only to adjust the AI's target. Set
-- as big as possible (though 25 is big enough) without AI's taking weird
-- routes.
local targetRadiusMultiplier = 5

-- Multiplier for AI aggression, higher means more race-like AI, but also tends
-- to make them crash.
local aiAggression = 
{
	race_Player1 = 5,
	race_Player2 = 5,
	race_Player3 = 5,
	race_Player4 = 5,
	race_Player5 = 5,
}

--------------------------------------------------------------------------------
-- AI code
--------------------------------------------------------------------------------
require('mathlib')

local lapCount = 1
local lapConfig = {}
local raceRoute = {}
local raceData = {}

-- Helper function to see if the car has finished
local function isFinished(car)
	return (raceData[car.name].raceLapIndex > lapCount)
end

-- Converts a vec3 to something I can compare to node positions.
local function vec2Point3f(v)
    return Point3F(v.x, v.y, v.z)
end

-- Determines the node closest to the given waypoint.
local function findClosestRoad(waypoint)
	local position = vec3(waypoint:getPosition())
	
	local bestRoad1 = nil
    local bestRoad2 = nil
    local bestDist = math.huge
	
    for nid, n in pairs(map.map.nodes) do
		local posX = vec3(n.pos)
		
        for lid, dif in pairs(n.links) do
			-- lid and dif are sometimes switched
			local m = map.map.nodes[lid] or map.map.nodes[dif]
			
			if m ~= nil then
				local posY = vec3(m.pos)
				local curDist = position:squaredDistanceToLineSegment(posX, posY)
				
				if curDist < bestDist then
					bestDist = curDist
					bestRoad1 = nid
					bestRoad2 = lid
				end
			end
        end
    end

    return bestRoad1, bestRoad2
end

-- Function the gets the route for the race
local function initRaceAI(scenario)

	-- Copy data first
	lapConfig = scenario.lapConfig
	lapCount = scenario.lapCount
	
	raceRoute = {}
	
	for node = 1, #scenario.lapConfig do
		local waypoint = scenario.lapConfig[node]
		
		-- Append the waypoint to the route
		table.insert(raceRoute, waypoint)
		
		-- Add a node to the map if it doesn't exist for this waypoint
		if not map.map.nodes[waypoint] then
			local wp = scenetree.findObject(waypoint)
			local wp1
			local wp2
			wp1, wp2 = findClosestRoad(wp)
			
			map.map.nodes[waypoint] =
			{
				pos = vec2Point3f(vec3(wp:getPosition())),
				radius = wp:getScale():len() * 0.5,
				links = { [wp1] = 1, [wp2] = 1 },
				segTags = {},
				manual = 1,
			}
		end
	end
end

-- Returns the next indices
local function getNextIndices(lapIndex, nodeIndex)
	nodeIndex = nodeIndex + 1
	
	if nodeIndex > #lapConfig then
		nodeIndex = 1
		lapIndex = lapIndex + 1
	end
	
	return lapIndex, nodeIndex
end

-- Sets up a specific car and starts the AI
local function startRaceAI(car)
	-- Set blank data for this car
	local name = car.name
	
	raceData[name] =
	{
		raceNodeIndex = 1,
		raceLapIndex = 1,
		targetNodeIndex = 1,
		targetLapIndex = 1,
		targetSet = false,
	}
	
	-- Set the ai to manual
	car:queueLuaCommand('ai.setState({mode="manual"})')
	
	-- Set ai aggression
	local aggression = aiAggression[name]
	car:queueLuaCommand('ai.setAggression(' .. aggression .. ')')
end

-- Updates the race AI, should be called every tick
local function updateRaceAI(car)
	if not car or not car.getPosition then
		return
	end
	
	local opponentPos = car.getPosition()
	local name = car.name
	local data = raceData[name]
	
	if not data then
		return
	end
	
	local targetWaypoint = raceRoute[data.targetNodeIndex]
	local targetNode = map.map.nodes[targetWaypoint]
	
	-- See if we should set the target
	if not data.targetSet and data.targetNodeIndex <= #raceRoute then
		car:queueLuaCommand('ai.setTarget("' .. targetWaypoint .. '")')
		data.targetSet = true
	end
	
	-- Get distances to target
	local targetDistance = math.abs((targetNode.pos - opponentPos):len())
	local targetRadius = targetNode.radius * targetRadiusMultiplier
	
	-- Update the car's target
	if targetDistance < targetRadius then
		data.targetLapIndex, data.targetNodeIndex = getNextIndices(data.targetLapIndex, data.targetNodeIndex)
		data.targetSet = false
	end
	
	-- See if we have touched a node in the race
	local raceWaypoint = raceRoute[data.raceNodeIndex]
	local raceNode = map.map.nodes[raceWaypoint]
	
	-- Get distances to race node
	local raceDistance = math.abs((raceNode.pos - opponentPos):len())
	local raceRadius = raceNode.radius * raceRadiusMultiplier
	
	-- Update the car's race position
	if raceDistance < raceRadius then
		data.raceLapIndex, data.raceNodeIndex = getNextIndices(data.raceLapIndex, data.raceNodeIndex)
	end
end

--------------------------------------------------------------------------------
-- Scenario code
--------------------------------------------------------------------------------
local M = {}

local opponentObjects = {}

local function fail(reason)
    scenarios.finish({failed = reason})
end

local function getNumberName(n)
	if n == 1 then
		return n .. 'st'
	elseif n == 2 then
		return n .. 'nd'
	elseif n == 3 then
		return n .. 'rd'
	else
		return n .. 'th'
	end
end

local function onScenarioChange(scenario)
	initRaceAI(scenario)

	for i, name in pairs(opponents) do
		opponentObjects[name] = scenetree.findObject(name)
		startRaceAI(opponentObjects[name])
	end
end

local function onRaceTick(raceTickTime)
	for i, name in pairs(opponents) do
		updateRaceAI(opponentObjects[name])
	end
end

local function onRaceResult()
	local place = 1
	
	for i, name in pairs(opponents) do
		if isFinished(opponentObjects[name]) then
			place = place + 1
		end
	end
	
	if place > 1 then
		fail('You finished in ' .. getNumberName(place) .. ' place!')
		return
	end
end

M.onScenarioChange = onScenarioChange
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult

return M